#!/usr/bin/env python
# coding: utf-8

# # `o()`: attr-access for `dict`s
# Wrap `dict`s (or kwarg lists) to allow direct `.`-accessing of items.
# 
# Before:
# ```python
# x={'a':1,'b':2}  # given some dict
# x['a'], x['b']   # access fields via "getitem" syntax
# ```
# After:
# ```python
# x=o({'a':1,'b':2})  # wrap dict in o()
# x.a, x.b            # access members via "getattr" syntax
# ```
# Or, instantiate directly with kwargs:
# ```python
# x=o(a=1,b=2)
# x.a, x.b
# ```
# 
# **Contents:**
# - [Implementation](#Implementation)
# - [Examples](#Examples)

# ## Implementation

# In[1]:


class o(dict):

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise ValueError(f'≤1 positional args required, got {len(args)}')
        
        if args:
            (data,) = args
            if type(data) is not dict:
                raise ValueError(f'Single-arg o() ctor call needs dict arg, not {type(data)}: {data}')
            if kwargs:
                raise ValueError(f'Positional dict arg is exclusive with kwargs: {data}, {kwargs}')
        else:
            data = kwargs

        K = '_data'
        if K in data:
            raise ValueError(f"Reserved key '{K}' found in 'data' dict: {data}")

        super().__init__(**data)

        for k, v in data.items():
            if isinstance(v, dict) and not isinstance(v, o): v = o(v)
            super(o, self).__setattr__(k, v)

        super(o, self).__setattr__(K, data)

    def __dict__(self): return self._data
    
    def __setattr__(self, k, v):
        if isinstance(v, dict) and not isinstance(v, o): v = o(v)
        self._data[k] = v

    def __getattr__(self, k):
        try:
            v = self._data[k]
            if isinstance(v, dict) and not isinstance(v, o): v = o(v)
            return v
        except KeyError:
            raise AttributeError(f'Key {k}')

    def get(self, k, default):
        if k in self:
            return self[k]
        else:
            return default

    def __call__(self, *keys, default=None):
        obj = self
        keys = list(keys)
        while keys:
            key = keys.pop(0)
            if key in obj:
                obj = obj[key]
            else:
                return default

        return obj

    def __getitem__(self, k):
        v = self._data[k]
        if isinstance(v, dict) and not isinstance(v, o): v = o(v)
        return v
    def __setitem__(self, k, v): self._data[k] = v
    def __contains__(self, k): return k in self._data
    
    def __str__(self): return str(self._data)
    def __repr__(self): return repr(self._data)

    def __iter__(self): return iter(self._data)
    def items(self): return self._data.items()

    def __eq__(self, r):
        if isinstance(r, o):
            return self._data == r._data
        if isinstance(r, dict):
            return self._data == r
        return NotImplemented

    def __ne__(self, r):
        if isinstance(r, o):
            return self._data != r._data
        if isinstance(r, dict):
            return self._data != r
        return NotImplemented

    def __hash__(self): return hash(self._data)


# ## Examples

# In[2]:


o1 = o(a=1,b=2)


# In[3]:


o1.a, o1.b


# In[4]:


from .context import catch
with catch(AttributeError): o1.c


# In[5]:


x={'c':3,'d':4}
o2 = o(x)
o2.c, o2.d


# In[6]:


x['e']=5
o2.e


# In[7]:


o2.c = 'ccc'
x


# In[8]:


o2


# In[9]:


str(o2)


# In[10]:


'c' in o2


# In[11]:


x


# In[12]:


list(iter(x))


# In[13]:


'z' in o2


# In[14]:


oo = o(a={'b':1})
oo


# In[15]:


oo.get('b', 'yay')


# In[16]:


oo.a.b




from argparse import ArgumentParser
from re import match
import shlex
import sys


class ArgsParser(ArgumentParser):
  '''ArgumentParser that adds an "--args" flag that takes extra arguments as a single Bash string.

  The value passed to "--args" will be split with `shlex`, and converted into potentially multiple arguments.

  Note that the value typically needs to be passed as "--args=<value>" (as opposed to as two Bash arguments: "--args", "<value>").

  Useful for e.g. specifying arguments across multiple YAML files that lack the ability to concatenate lists.
  '''

  ARGS_HELP = '''Additional arguments, passed as one bash-token string (e.g. "-f -x").

  If the string being passed begins with a "-" character (as it typically will), then this argument should be expressed as one bash token like "--args=-f -x" (as opposed to two tokens: "--args", "-f -x"), so that argparse doesn\'t get confused.

  Useful for e.g. specifying arguments across multiple YAML files that lack the ability to concatenate lists.
  '''

  def __init__(self, *args, log=None, **kwargs):
    super(ArgsParser, self).__init__(*args, **kwargs)
    self.add_argument('--args', help=self.ARGS_HELP)
    if log is True:
      self._log = print
    elif log:
      if hasattr(log, 'write'):
        def _print(msg):
          log.write(msg + '\n')
        self._log = _print
      else:
        self._log = log
    else:
      self._log = None

  def log(self, *args):
    if self._log:
      self._log(*args)

  def parse_args(self, *args, **kwargs):
    extra_args, _ = self.parse_known_args()
    self.log(f'extra_args: {extra_args}')

    extra_args = extra_args.args
    if extra_args:
      extra_args = shlex.split(extra_args)
      self.log(f'post-shlex --args: {extra_args}')
      [(idx, args_arg)] = [
        (idx, arg)
        for idx, arg in enumerate(sys.argv)
        if arg.startswith('--args')
      ]
      args_arg_span_size = 1 if args_arg.startswith('--args=') else 2

      args = sys.argv

      # Strip 'python' executable from front of sys.argv, if present:
      if match('python3?', args[0]):
        args = args[1:]

      # Strip called python file from sys.argv; ArgumentParser.parse_args() does this on its own
      # when called with no arguments, but we are passing explicit arguments here, so need to
      # mimic it.
      args = args[1:]

      # Strip "--args=<value>" or ["--args","<value>"] args:
      args = sys.argv[1:idx] + extra_args + sys.argv[(idx+args_arg_span_size):]

      args = super(ArgsParser, self).parse_args(args)
    else:
      args = super(ArgsParser, self).parse_args()

    return args

  def exit(self, status=0, message=None):
    if status:
      args_msg = 'error: argument --args: expected one argument'
      if args_msg in message:
        raise Exception(f'argparse is exiting with message "{args_msg}". Most likely, you are passing a value to the "--args" flag as a separate Bash argument, and that value begins with a dash ("-"), confusing argparse. Pass "--args=<value>" as one bash token instead.')
      else:
        raise Exception(f'argparse error: {message}')
    exit(status)



#!/usr/bin/env python
# coding: utf-8

# # imports
# Common Jupyter imports and helpers to wildcard import from other notebooks

# Helper for optional imports:

# In[7]:


from contextlib import suppress
_try = suppress(ImportError, ModuleNotFoundError)


# ## stdlib
# Common imports (and associated helpers) from the Python standard library:

# ### Date/Time

# In[ ]:


from dateutil.parser import parse
from datetime import datetime as dt, date
with _try: from pytz import UTC
now = dt.now()
ISO_DATE_FMT = '%Y-%m-%dT%H:%M:%S'
runtime = now.strftime(ISO_DATE_FMT)
today = now.strftime('%Y-%m-%d')


# ### Paths

# In[ ]:


from pathlib import Path
def mkdir(path, *args, exist_ok=True, **kwargs):
    os.makedirs(str(path), *args, exist_ok=exist_ok, **kwargs)
    return path

def mkpar(path, *args, **kwargs):
    if isinstance(path, Path):
        path.parent.mkdir(exist_ok=True, parents=True)
    else:
        mkdir(dirname(path))
    return path


# ### Other

# In[ ]:

from configparser import ConfigParser
from dataclasses import dataclass

import functools
try:
    # Python 3.8
    from functools import cached_property, singledispatchmethod
except ImportError:
    try:
        # Python ≤3.7; pip install cached-property
        from cached_property import cached_property
    except ImportError as e:
        pass

from functools import partial, lru_cache, namedtuple, reduce, singledispatch

from glob import glob

from itertools import combinations, combinations_with_replacement, permutations

import json

import os
from os import cpu_count, environ as env, getcwd
from os.path import basename, dirname, exists, isabs, isdir, isfile, islink, join, sep, splitext

import re
from re import match, fullmatch, IGNORECASE, MULTILINE, DOTALL, search, split, sub

import shlex
from shutil import copy, copyfileobj, move, rmtree

from subprocess import check_call, check_output, CalledProcessError, DEVNULL, PIPE, Popen

import sys
from sys import stdout, stderr, executable as python, exit, platform

from tempfile import NamedTemporaryFile, TemporaryDirectory, TemporaryFile
from time import sleep


# ## Sibling modules
# Some other notebooks and Python files from this repo:

# In[3]:


from . import process
from .process import *

from . import pnds
from .pnds import *

from .cd import cd
from .o import o

from .git import Git, Repo, make_repo

from .args_parser import *

from .context import *

from .collections import coerce, singleton


# ## Optional Modules

# In[8]:


# joblib: easy parallelization
with _try:
    from joblib import Parallel, delayed
    parallel = Parallel(n_jobs=cpu_count())

# yaml
with _try:
    import yaml
    # Fix a bad default in PyYAML (cf. https://github.com/yaml/pyyaml/issues/110)
    yaml.dump = partial(yaml.dump, sort_keys=False)
    yaml.safe_dump = partial(yaml.safe_dump, sort_keys=False)

# requests
with _try:
    from requests import           get as   GET,          post as  POST,           put as   PUT,         patch as PATCH


# ## PyData / Scientific Python

# In[9]:


import numpy as np
from numpy import concatenate, array, ndarray, matrix, nan


# In[10]:


with _try: import seaborn as sns
with _try: import matplotlib.pyplot as plt
with _try: from scipy.sparse import spmatrix, coo_matrix, csr_matrix, csc_matrix




#!/usr/bin/env python
# coding: utf-8

# # Pandas imports / aliases / helpers

# In[ ]:


from dateutil.parser import parse
from pathlib import Path

from numpy import nan, array, ndarray
from os import remove
from os.path import exists, isdir, splitext

import pandas as pd
from pandas import     concat, DataFrame as DF, Series,     isna,     read_csv, read_excel, read_json, read_parquet, read_sql, read_sql_query, read_sql_table,     date_range, to_datetime as to_dt, Timedelta as Δ, NaT

from shutil import rmtree

# ## Config

# In[ ]:


def display(r=None, c=None):
    '''Set the default number of rows and columns for Pandas to display'''
    if r:
        pd.options.display.max_rows = r
    if c:
        pd.options.display.max_columns = c

display(100, 100)


# ## Concat / Creation

# In[ ]:


def sxs(*dfs, **kwargs):
    '''Concat some DataFrames "side by side"'''
    return concat(dfs, axis=1, **kwargs)


# ## Counting

# In[ ]:


def series_counts(series, ascending = None):
    df_new = DF(series.value_counts())
    column_name = df_new.columns[0]
    df_new['index'] = df_new.index
    ascending = [False,True] if ascending is None else ascending
    return df_new.sort_values(by = [column_name,'index'], ascending = ascending)[column_name]


# In[ ]:


def sort_by_series(df, series):
    tmp_col_name = series.name or '_'
    while tmp_col_name in df:
        tmp_col_name=f'_{tmp_col_name}'

    if isinstance(series, str):
        series = df[series]

    df2 = df.copy()
    df2[tmp_col_name] = series
    return         df2         .sort_values(tmp_col_name)         .drop(columns=tmp_col_name)


# ## File paths

# In[ ]:


def file_dict(path):
    name = path.name
    pieces = name.rsplit('.', 1)
    if len(pieces) == 2:
        base, xtn = pieces
    else:
        [ base ] = pieces
        xtn = nan
    
    return dict(parent=path.parent, name=name, base=base, xtn=xtn, exists=path.exists())


def annotate_files(files):
    if files.empty:
        dtypes = {
            'parent': object,
            'name': str,
            'base': str,
            'xtn': str,
            'exists': bool,
        }
        return DF([], index=files, columns=dtypes.keys()).astype(dtypes)
    else:
        extra = files.apply(file_dict).apply(Series)
        files = sxs(files, extra).set_index(files.name)
        return files


def load_files(files, glob=None, name=None):
    if isinstance(files, Path):
        name = name or 'path'
        files = Series(files.glob(glob), name=name, dtype=object)
    else:
        name = name or files.name
    files = annotate_files(files)
    return files


# ## Dates

# In[ ]:


def expand_date(s):
    d = parse(s).date()
    return dict(date=d, year=d.year, month=d.month, day=d.day)


def to_parquet(df, out_path, verify_extension=True, *args, **kwargs):
    if verify_extension:
        _, extension = splitext(out_path)
        if extension != '.parquet':
            raise Exception(f"Refusing to write parquet dataset to non-parquet path {out_path}")

    if exists(out_path):
        if isdir(out_path):
            rmtree(out_path)
        else:
            remove(out_path)

    return df.to_parquet(out_path, *args, **kwargs)
